--[[ Many Minor Tweaks (MMT): Revamped 
Author: birbhorse aka EeveeEuphoria
Additional code provided by: lua-users.org, Monster ɪestyn (SRB2 Discord), Tatsuru (SRB2 Discord), SRB2 Wiki, Sonic Team Jr.
Additional credits: customsave_io.lua (Serialize and Deserialize functions) by sprkizard AKA szv (found here: https://github.com/sprkizard/srb2-lua-lib/blob/master/game/customsave_io.lua)
Additional credits: PBT3.lua by amberbee/ampersand

This is an entire from-the-ground-up restructuring of Many Minor Tweaks, built to eliminate many of the flaws from the original version.
Almost all the commands have been moved to be used in the HUD instead.

                 __     __            __     __       ________
                /  \   /  \          /  \   /  \     |___  ___|
               /    \_/    \        /    \_/    \       |  |
              /  /\     /\  \      /  /\     /\  \      |  |
             /  /  \   /  \  \    /  /  \   /  \  \     |  |
            /__/    \_/    \__\  /__/    \_/    \__\    |__|


]]

--[[ CHANGELOG:

	[*] Added a mod system functionality to MMT! This allows other mods to use MMT's menu system to have their options display here!
	How this works is that there's two functions given, MMT_AddModToMMT and MMT_AddModSettings, which will work with multiple kinds of mods. An example mod is provided! There is also a .txt file in the PK3 itself, documenting how to use this!
	One of my other mods, "SRB2: Chaos Edition", will been updated real soon to support this!
	[*] New option added: "SFX Style"! This allows you to change the sounds that are played in-game.
	This is a single-player only option, due to replacing sounds not working properly in netgames.
	[*] Added new misc. options:
	- Preserve Shield Between Zones
	- Recurl Multiple Times
	[*] A changelog for MMT itself has been added to the menu!
	[*] If a player is dead, the MMT menu deactivates itself.
	[*] "Total Rings" now also displays how much of a deficit you have in terms of rings. It'll also show you if you've collected more rings than you should've been able to!
	[*] "Ring Special Stage Entrance" now works when exiting a special stage too!
	[*] Removed "sfx_tokrin", because it's apparantly SFX that originated from Sonic CD all this time!
	[*] Fixed this mod being incompatible with other mods that use the PBT3 system (i.e. the Shadow mod by Lach)
	The reason why it was breaking was due to MMT including a slightly modified version of PBT3, and I didn't change the proper things around to prevent conflicts.
	[*] Fixed "Recurl" potentially making on-ground behavior act weird.
	[*] Fixed "Tails Ball" allowing tails to carry players while spinning.
	[*] Fixed "Tails Ball" in "bop" mode not playing the correct sounds if you were using a custom skin with custom sounds.
	[*] Fixed sub-menus not properly saving what page the player was on.
	[*] Fixed sub-menus starting on the last page.
	[*] Fixed "Deaths Counter" counting two deaths instead of one if the player died less than 2 seconds after respawning.
]]

--[[ TODO LIST:
KEY:	[ ] = nothing done
		[?] = partially done/untested
		[x] = complete
		{ } = high priority, must complete before release
	
	- [?] option to change various sfx
		- [ ] spinning sign noise
			- probably can only do in single player, i mean, this already is only single player, but this really can't work in multiplayer
		- [ ] "pop" sound, monitors & enemies
			- probably need to do similar thing to springs
			- [ ] sfx_kc2e knuckles' chaotix
	
	- [ ] apply recurl when in air?
		- would remove the need to figure out what state the player is in; it'd only check for the applied jump state
	
	- [ ] create way of setting variable back to default
		- [ ] hold down confirmation key for a few seconds when entering the menu to reset to default
			- [ ] make it show a timer overlay, indicating setting will be reset if the player keeps holding down the key
		- [ ] have "reset to defaults" option too 

	- [x] give up
		- [x] disable the hud for netgames
		- [x] warn the player to use "mmtchangevar" when the menu key is pressed 
	
	- [ ] implement secret pizza menu into the HUD
		- [ ] do it by trying to scroll down on the final menu item 50 times (which can be done by just holding down)
	
	- [ ] fix hud in multiplayer
		- figure out how to set up a netgame that actually gets affected by this, local instances aren't affected
		- potential fix: seperating out the changes to the player variable
			- use a variable (i.e. player.MMT.ChangeThisSetting) and have the main MMT script work with it instead
			- or use a regular thinkframe after the fact to do it
	
	- [ ] allow the inverse for mmtchangevar, i.e. "mmtchangevar mmtmenukey off"
		- detect it in mmtchangevar list, if it is there, reverse the value back to normal so the rest of the script doesn't freak out 
	
	- [ ] color-blind accessibility, for mmtchangevar 
		- have non-selectable options be presented more clearly than just a color
			- [ ] replace * with x ?
		- also great for dedicated server hosts, because the command line is boringly monochrome
	
	- [ ] "MMTdebuggerprinting" isn't syncing properly in multiplayer? why?
		- fixing this might fix the damn problem with everything not syncing in the first place
		- [x] TEST THE BEHAVIOR WITH mmtchangevar
			- it actually functions with this???? that definitely means the HUD is messing things up with netgames 

	- [ ] things to do when music detection is possible:
		- [ ] fix bug with super music, if invincibility is playing when player turns super, the music doesn't start playing again
		- [ ] make super music take effect immediately, regardless if player is super or not right now

	- [ ] when saving local settings in multiplayer, it makes everyone crash because it's trying to write to everyone else's stuff????
		- [x] disable it for now i guess, fix it later
			- disabled savesettings stuff in lua io code

	- [ ] when fake shield is out, render the effects shields make (e.g. the double-hit shield's stopping effect)
		- only relevant for invincibility
		- [ ] check the shield's raisestate?
			- check P_MobjSceneryThink in p_mobj.c, it has a bunch of thinker logic for the shield orbs  
	
	- [ ] there's a merge request for P_DoSuperTransformation to allow the super jingle to be disabled; keep an eye on it

	- [ ] multiplayer-based emblems
		- they won't actually add to the emblem count of course, but they can be collected just like normal in multiplayer 
		- would need to do the opposite for SPOnly things 
	
	- [ ] allow other players names to be shown at all times, for co-op only
		- gets gradually more transparent the further away they are, name is always shown right above the player's height
		- R_GetScreenCoords?
		- R_ScreenTransform?
			- wait for someone to get back to me on fixed code for that 
				- they never did :(
	
	- [ ] spin state isn't properly applied on chrispychars after doing a double-jump w/ re-roll?
		- [ ] don't allow re-curl when that double jump occurs
			- [ ] unfortunately impossible to determine that unless i use extremely convoluted work arounds >:( why does their code unset PF_THOKKED?!?! it's used to detect if a character used their ability!!!!
	- [ ] tailsball works on fluffy tails, but it doesn't affect the flight timer
		- it uses the standard player.powers[pw_tailsfly], but something in it is changing it after my script
		- it's more broken than i thought actually, it allows you do do re-flight any time after curling up; and it doesn't let you re-curl afterwards
	
	- [ ] afk timer
		- toggleafk should be able to be deactivated in multiplayer.
		
	- [?] classicspecialstage
		- [x] local variable, single-player only too
		- [?] two options: "instant warp token" and "50 rings"
			- [ ] instant token transports the player to a special stage, ala sonic 3&K
				- save player's x, y, z, rings, & level time position, so it can be re-applied when the special stage is done
				- use G_SetCustomExitVars to both send the player to the stage, and to have the player return 
				- would also need to keep track of what tokens were collected; create a list of token mobjs collected, and delete them upon re-entry to the level
					- could also do this for all the rings and enemies and whatnot
				- use the sonic 1 proto effect when entering a special stage ring, that'd be neato
			- [x] 50 rings gives the player a token at the end of the stage, albeit quietly 
	
	- [ ] fix the position of totalrings in multiplayer

	- [ ] VS_MPSpecialStagesinSP-v1.soc, but in lua form
		- would require changing the level exit for every emerald token obtained, hopefully that's do-able?
		- if this is possible, do something where it's check-point based special stages; incorporate it into classicspecialstage
		- also allow for the inverse
		- not possible yet! in the next version of SRB2 it will be though, i think 
			- G_SetCustomExitVars(nextmap)
			- this doesn't override special stages yet, but in the next version it will... i think?
	
	- improvements:
		- [ ] "Super music should take effect immediately./Super music should be main level's theme when disabled and should be able to interrupt the invincibility jingle only when supermusic is disabled."
			- can't do yet, need to wait for the function to detect currently playing music to happen 
]]

--- [[[ Player Think ]]] ---

addHook("PlayerThink", function(player) if player.bot > 0 return end if player.mo == nil return end	
	-- [[ Auto Load / Auto-Load ]] --
	if not player.MMT.LoadedSettings
		local Retry = false
		player.MMT.loadtimer = $ + 1
		if player.MMT.loadtimer > (TICRATE/4)
			local filez2 = io.openlocal("client/" .. MMTfolder .. "/" .. "MMTautoload" .. ".dat", "r")
			if filez2
				local output = filez2:read("*all")
				--mmtdebuggerPrint("output: " +output)
				if output != nil and output == "on"
					MMT_LoadSettings(player)
				else -- manually set autoload to off 
					player.MMT.MMTautoload = "off"
				end
				filez2:close()
			else
				MMT_SaveSpecificSetting(player, "MMTautoload", true) -- creates a new file
				Retry = true -- doesn't load settings yet, since that file was just created
			end
		if not Retry player.MMT.LoadedSettings = true end
		end
	end
	
	-- [[ Player Functions ]] --

	local function TailsBallTime()
		if player.MMT.MMTtailsball != "off"
		and player.bot != 1
		and (player.charability == CA_FLY) and (player.charability2 == CA2_SPINDASH) and (player.panim == PA_ABILITY) and (skins[player.mo.skin].name != "eggman")
		and (player.pflags & PF_USEDOWN)
		and not player.MMT.waittogetdown
			return true
		else
			return false
		end
	end
	
	local function SouperSoup()
		player.MMT.JustActivatedSuper = true
		player.MMT.waittogetdown = true
		player.pflags = $ & ~PF_JUMPED
		player.charflags = $1 | SF_SUPER
		if player.MMT.MMTsupermusic == "on"
			--MMT_DoSuperTransformation(player, false)
			P_DoSuperTransformation(player, false)
			mmtdebuggerPrint("Regular Superkey Transformation go!")
		else
			player.powers[pw_super] = 1
			player.mo.state = S_PLAY_SUPER_TRANS1
			S_StartSound(player.mo, sfx_supert, player)
			mmtdebuggerPrint("supermusic disabled, Superkey transformation go!")
		end
	end

	-- [[ Logic ]] --
	
	-- [ Tails Ball ] --
	if TailsBallTime() 
		player.MMT.waittogetdown = true
		player.MMT.boptime = false
		if (player.MMT.MMTtailsball == "jump" or player.MMT.MMTtailsball == "bop")
			S_StartSound(player.mo,sfx_3db16)
			player.mo.state = S_PLAY_JUMP
			player.pflags = $1 | PF_JUMPED
			player.pflags = $1 | PF_THOKKED
			player.pflags = $ & ~PF_CANCARRY
			--player.MMT.MMTtailsballsfxtimer = 30
			player.MMT.istailsball = true
			player.MMT.initialfall = false
			player.MMT.beginthebopping = false
			player.MMT.allowreflight = false
			player.MMT.usedreflight = false
		elseif (player.MMT.MMTtailsball == "spin")
			S_StartSound(player.mo,sfx_3db16)
			player.mo.state = S_PLAY_ROLL
			player.pflags = $1 | PF_SPINNING
			player.pflags = $1 | PF_THOKKED
			player.pflags = $ & ~PF_CANCARRY
			--player.MMT.MMTtailsballsfxtimer = 30
			player.MMT.istailsball = true
		end
	end
	-- [ tailsball bop ] --
	if player.MMT.istailsball and (player.MMT.MMTtailsball == "bop") and not P_IsObjectOnGround(player.mo)
		if player.MMT.istailsball and not player.MMT.beginthebopping and not player.MMT.initialfall
			mmtdebuggerPrint("INITIAL FALL: " .. player.mo.momz)
			player.MMT.initialfall = true
		end
		if player.MMT.initialfall and ((P_MobjFlip(player.mo) == 1 and player.mo.momz < 0) or (P_MobjFlip(player.mo) == -1 and player.mo.momz > 0))
			player.MMT.beginthebopping = true
			player.MMT.initialfall = false
			mmtdebuggerPrint("THE BOPPENING CAN HAPPEN: " .. player.mo.momz)
		end
		if player.MMT.beginthebopping and ((P_MobjFlip(player.mo) == 1 and player.mo.momz > 0) or (P_MobjFlip(player.mo) == -1 and player.mo.momz < 0))
		and not player.MMT.allowreflight and not player.MMT.usedreflight and player.panim == PA_JUMP
			player.MMT.allowreflight = true
			mmtdebuggerPrint("ALLOWING REFLIGHT: " .. player.mo.momz)
		end
		if player.MMT.allowreflight and (player.cmd.buttons & BT_JUMP) and not player.MMT.usedreflight
			player.MMT.usedreflight = true -- set to false in tailsball main script portion to make it happen more than once
			player.mo.state = S_PLAY_FLY
			player.powers[pw_tailsfly] = TICRATE
			player.MMT.waittogetdown = false
			mmtdebuggerPrint("USED REFLIGHT!")
		end
		if player.MMT.usedreflight and player.mo.state == S_PLAY_FLY
			if (player.MMT.MMTchangesfx == "sonic cd" or player.MMT.MMTchangesfx == "sonic 1" or player.MMT.MMTchangesfx == "sonic 2" or player.MMT.MMTchangesfx == "sonic 3k")
				if not S_SoundPlaying(player.mo, sfx_s3kba)
					S_StartSound(player.mo, sfx_s3kba, player)
				end
			else
				if not S_SoundPlaying(player.mo, skins[player.mo.skin].soundsid[SKSPUTPUT])
					S_StartSound(player.mo, skins[player.mo.skin].soundsid[SKSPUTPUT], player)
				end
			end
		end
	end
	
	-- [[ Super Key and Super Music ]] --
	if ((player.MMT.MMTsuperkey != "off") or (player.MMT.MMTsupermusic == "off")) and All7Emeralds(emeralds) and player.rings >= 50 and not player.powers[pw_super] and player.MMT.MMTclassicsuper == "off" and MMT_SoupTime(player)
	and not P_IsObjectOnGround(player.mo) and ((player.MMT.MMTsuperkey != "off" and (player.cmd.buttons & MMT_KeyToNum(player.MMT.MMTsuperkey))) or (player.MMT.MMTsuperkey == "off" and player.MMT.MMTsupermusic == "off" and player.cmd.buttons & BT_USE))
	and ((player.powers[pw_shield] == 0) or (player.powers[pw_shield] > 0 and MMT_KeyToNum(player.MMT.MMTsuperkey) != BT_USE))
	and ((MMTeverysuper == "on") or skins[player.mo.skin].flags & SF_SUPER)
		SouperSoup()
	end
	
	-- [ super sector ] -- 
	if ((player.MMT.MMTsuperkey != "off") or (player.MMT.MMTsupermusic == "off")) and All7Emeralds(emeralds) and not player.powers[pw_super] and player.MMT.MMTclassicsuper == "off"
	and P_PlayerTouchingSectorSpecial(player, 4, 6)
	and ((MMTeverysuper == "on") or skins[player.mo.skin].flags & SF_SUPER)
		mmtdebuggerPrint("special sector activated super form!")
		player.rings = 50
		SouperSoup()
	end
	
	
	-- [[ Cancel Super ]] --
	if player.MMT.MMTcancelsuper != "off" and not P_IsObjectOnGround(player.mo) and player.powers[pw_super] and player.MMT.MMTclassicsuper == "off" and MMT_SoupTime(player) 
	and (player.cmd.buttons & MMT_KeyToNum(player.MMT.MMTcancelsuper))
	and (abs(player.mo.momz) > FRACUNIT/18)
		player.pflags = $1 | PF_THOKKED
		player.MMT.RestoreThatShield = player.powers[pw_shield]
		P_SpawnShieldOrb(player)
		player.powers[pw_flashing] = TICRATE
		player.powers[pw_super] = 0
		player.powers[pw_shield] = player.MMT.RestoreThatShield
		player.mo.color = player.skincolor
		player.MMT.waittogetdown = true
		P_RestoreMusic(player)
		mmtdebuggerPrint("Super has been cancelled!")
	end
	
	-- [[ Classic Super ]] --
	if player.MMT.MMTclassicsuper == "on" and not P_IsObjectOnGround(player.mo) and MMT_SoupTime(player) and (player.cmd.buttons & BT_JUMP == 0)
	and player.rings >= 50 and not player.powers[pw_super]
	and ((player.powers[pw_shield] == 0) or (player.powers[pw_shield] > 0 and KeyToNum(player.MMT.MMTsuperkey) != BT_USE)) 
	and (player.pflags & ~PF_JUMPDOWN)
		SouperSoup()
	end
	
	-- [ Replace Super Forms ] --
	if ((player.MMT.MMTsuperkey != "off") or (player.MMT.MMTsupermusic == "off"))
		if not player.powers[pw_super] -- won't apply to player if they're already super 
			player.charflags = $ & ~SF_SUPER
		end
	elseif MMTeverysuper == "off"
		if not player.powers[pw_super]
			if skins[player.mo.skin].flags & SF_SUPER
				player.charflags = $1 | SF_SUPER
			else
				player.charflags =  $ & ~SF_SUPER
			end
		end
	elseif MMTeverysuper == "on"
		player.charflags = $1 | SF_SUPER
	elseif MMTeverysuper == "disable"
		player.charflags =  $ & ~SF_SUPER
		player.powers[pw_super] = 0
	end
	
	-- [[ Music Fade ]] --
	if player.MMT.MMTmusicfade == "on" and G_EnoughPlayersFinished()
		S_SetInternalMusicVolume(player.MMT.MusicVolume, player)
		player.MMT.beginfade = true
	end
	
	-- [[ Instant Respawn ]] --
	if player.MMT.MMTinstantrespawn == "on"
	and (gametype == GT_COOP) and player.playerstate == PST_DEAD
		mmtdebuggerPrint("Instant Respawn!")
		player.playerstate = PST_REBORN
	end
	
	-- [[ Preserve Shield ]] --
	if MMTpreserveshield == "on"
		if (player.pflags & PF_FINISHED) and not player.MMT.ShieldAlreadySaved and not G_IsSpecialStage(gamemap)
			player.MMT.ShieldAfterLevel = player.powers[pw_shield]
			player.MMT.ShieldAlreadySaved = true
			mmtdebuggerPrint("Saving shield! Shield saved is: " +player.MMT.ShieldAfterLevel)
		end
	end

	if player.MMT.ShieldAlreadySaved and not G_IsSpecialStage(gamemap)
		player.MMT.ShieldLevelName = mapheaderinfo[gamemap].lvlttl
	end

	-- [[ Super Trail ]] --

	if player.MMT.MMTsupertrail == "on"
	and player.powers[pw_super] and not player.mo.hyperflashcolor -- prevents conflict with hyper abilities
	and not (player.speed < 4*FRACUNIT and player.mo.momz == 0)
		local traily = P_SpawnGhostMobj(player.mo)
		traily.fuse = 4
		traily.rollangle = player.mo.rollangle
		if (traily.tracer)
			traily.tracer.fuse = 4
		end
	end
	
	-- [[ Super Shield ]] --
	if player.MMT.MMTsupershield == "on" and player.bot != 1
		if not player.MMT.FakeShield and player.playerstate == PST_LIVE
			player.MMT.FakeShield = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_FAKE_SHIELD)
			player.MMT.FakeShield.target = player.mo
		end
		if player.MMT.FakeShield and player.MMT.FakeShield.valid
		local shield = player.MMT.FakeShield
			if (player.powers[pw_super] or player.powers[pw_invulnerability]) and shield.DontRender
				mmtdebuggerPrint("player is super, re-enabling render")
				shield.DontRender = false
				shield.SetState = false
			end
		end
		
		if player.playerstate == PST_DEAD and player.MMT.FakeShield and player.MMT.FakeShield.valid
			player.MMT.FakeShield.TerminateSelf = true
			player.MMT.FakeShield = nil 
		end
	end

	if player.MMT.MMTsupershield == "off"
		if player.MMT.FakeShield
			player.MMT.FakeShield.TerminateSelf = true
			player.MMT.FakeShield = nil
		end
	end
	
	--- [[ Re Curl / Recurl ]] --
	if (player.MMT.MMTrecurl != "off")  and (player.charability2 == CA2_SPINDASH) and (skins[player.mo.skin].name != "eggman") and not G_IsSpecialStage(gamemap)
		if player.mo.sprite2 == SPR2_SPNG
			player.MMT.UsedSpring = true
			mmtdebuggerPrint("used spring")
		elseif (player.panim == PA_PAIN and not P_PlayerInPain(player))
		--or player.panim == PA_PAIN and not player.powers[pw_flashing]
			player.MMT.UsedTornado = true
			mmtdebuggerPrint("used tornado")
		elseif not player.MMT.UsedSpring and not player.MMT.UsedTornado and player.panim == PA_FALL and not (player.pflags & PF_SHIELDABILITY) and not (player.charability == CA_FLOAT and (player.pflags & PF_THOKKED))
			player.MMT.UsedMisc = true
			mmtdebuggerPrint("used misc")
		end
		
		if player.MMT.MMTrecurl == "spin"
			player.MMT.MMTrecurlKey = BT_USE
		else
			player.MMT.MMTrecurlKey = BT_JUMP
		end
		if not P_IsObjectOnGround(player.mo)
		and ((player.MMT.UsedSpring or player.MMT.UsedMisc) or (player.MMT.UsedTornado and (player.panim == PA_FALL)))
		and (player.cmd.buttons & player.MMT.MMTrecurlKey)
		and (player.charability2 == CA2_SPINDASH)
		and (not player.MMT.recurlwait or (player.MMT.UsedSpring and player.MMT.UsedTornado and player.panim == PA_FALL))
		and player.mo.state != S_PLAY_JUMP
			player.mo.state = S_PLAY_JUMP
			player.pflags = $1 | PF_JUMPED
			player.pflags = $1 | PF_THOKKED
			S_StartSound(player.mo, sfx_spin, player)
			if player.MMT.MMTrecurlmulti != "on" player.MMT.recurlwait = true end
			mmtdebuggerPrint("spin curl")
		end
	end

	-- [[ Token to Rings ]] --
	if player.MMT.TokenRingCollected
		P_GivePlayerRings(player, 50)
		P_AddPlayerScore(player, 1000)
		S_StartSound(player.mo, sfx_cdfm66)
		player.MMT.TokenRingCollected = false
	end
	
	-- [[ Player on Ground ]] -- 
	if P_IsObjectOnGround(player.mo)
		--player.MMT.HasGlided = false
		--player.MMT.PreviousFrameThokDone = false
		player.MMT.UsedSpring = false
		player.MMT.UsedMisc = false
		player.MMT.UsedTornado = false
		player.MMT.recurlwait = false
		--player.MMT.JustActivatedSuper = false
		-- rest is used for tailsball 
		player.MMT.waittogetdown = false -- used for multiple purposes too 
		player.MMT.initialfall = false
		player.MMT.PFTimer = 0
		player.MMT.beginthebopping = false
		player.MMT.istailsball = false
		player.MMT.allowreflight = false
		player.MMT.usedreflight = false
	end
	
	-- [[ Music Volume Logic ]] --	
	
	if player.MMT.beginfade or player.MMT.MenuActive 
		S_SetInternalMusicVolume(player.MMT.MusicVolume, player)
		if player.MMT.beginfade
			if player.MMT.MusicVolume > 0
				player.MMT.MusicVolume = $ - 1
			end
		elseif player.MMT.MenuActive 
			if player.MMT.MusicVolume > 66
				player.MMT.MusicVolume = $ - 2
			end
		end
	else
		if player.MMT.MusicVolume < 100
			player.MMT.MusicVolume = $ + 1
			S_SetInternalMusicVolume(player.MMT.MusicVolume, player)
		end
		
		if player.MMT.MusicVolume > 100
			player.MMT.MusicVolume = 100
			S_SetInternalMusicVolume(player.MMT.MusicVolume, player)
		end
	end
	
	if (player.pflags & PF_FINISHED) and not G_IsSpecialStage(gamemap) and MMTscorecontinues == "on"
		player.MMT.scoretocalculate = player.score
		--mmtdebuggerPrint("gamer")
	end
	player.MMT.skincont = player.mo.skin -- have to do this because intermission hud can't access mobj data, thanks stjr >:(
	player.MMT.colorcont = player.mo.color -- have to do this because intermission hud can't access mobj data, thanks stjr >:(
	
	-- [[ Classic Special Stage ]] --
	
	if not multiplayer and MMTclassicspecialstage == "50 rings, no tokens" or MMTclassicspecialstage == "50 rings, keep tokens"
		if player.rings >= 50 and (player.pflags & PF_FINISHED) and not All7Emeralds(emeralds) and not player.MMT.EnteredCS and not G_IsSpecialStage(gamemap)
			player.MMT.EnteredCS = true
			--S_StopSoundByID(player.mo, sfx_token)
			local token = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_TOKEN)
			token.GottaGo = true
			mmtdebuggerPrint("classic special stage time")
		end
	end
	
	if not player.MMT.SSE and player.MMT.MMTspecialstageentrance == "on"  and player.exiting > 0 and ((token > 0  and (player.pflags & PF_FINISHED)) or (G_IsSpecialStage(gamemap) and player.exiting < 43))
		player.MMT.SSE = true
		P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_MMTSPECIALSTAGEENT)
		S_StartSound(player.mo, sfx_s3kb3)
	end
	
	if player.MMT.SSE and not (mapheaderinfo[gamemap].typeoflevel & TOL_NIGHTS)
		player.mo.momx = 0
		player.mo.momy = 0
		player.mo.momz = 0
		player.mo.state = S_PLAY_STND
	end
	
	
	if G_IsSpecialStage(gamemap) 
		player.MMT.ContinueHudDontRender = true
	end 
	
	-- [[ Ban Skin ]] --
	
	if MMTbanskinNum > 1 and multiplayer and not MMT_IsPlayerAdmin(player)
		if MMT_invtablesearch(MMTbanskin, player.mo.skin) 
			player.powers[pw_nocontrol] = 1 -- todo: set in postthinkframe to allow no control, besides the menu controls 
			player.mo.momx = 0
			player.mo.momy = 0
			--player.mo.momz = 0
			player.pflags = $1 | PF_THOKKED
			player.powers[pw_flashing] = 5
			player.mo.state = S_PLAY_PAIN
			player.tossdelay = TICRATE
			player.weapondelay = TICRATE
			player.pflags = $1 | PF_INVIS
			if not player.MMT.BanSkinNotify
				player.MMT.BanSkinNotify = true
				S_StartSound(player.mo, sfx_s3kb2)
			end
		end
	end
	
	if multiplayer and not MMT_invtablesearch(MMTbanskin, player.mo.skin)
		player.MMT.BanSkinNotify = false
		player.pflags = $ & ~PF_INVIS
	elseif MMT_IsPlayerAdmin(player) -- if player is promoted to admin while wearing a banned skin, this disables the HUD
		player.MMT.BanSkinNotify = false
		player.pflags = $ & ~PF_INVIS
	end
	
	-- [[ Option Was Changed, KEEP AT BOTTOM OF PLAYERTHINK!]] --
	
	if player.MMT.MMTOptionChanged
		if player.MMT.MMTmenukey == "off"
			player.MMT.MenuActive = false
		end
		player.MMT.MMTOptionChanged = false
	end
	
	if player.MMT.MMTModOptionChanged
		player.MMT.MMTModOptionChanged = false
	end
	
end)

addHook("PostThinkFrame", function(thonk)
	for player in players.iterate do
		-- [[ Classic Special Stage ]] --
		if player.MMT.SSE
			S_StopSoundByID(player.mo, sfx_token) -- disables the sound when leaving a stage
		end
		-- [[ Change Various SFX ]] --
		if not netgame
			if player.MMT.MMTchangesfx != "off"
				if player.MMT.MMTchangesfx == "sonic cd"
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSJUMP], sfx_cdfm02)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSPUTPUT], sfx_s3kba)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSPUDPUD], sfx_s3kbb)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSSPIN], false)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSSPNDSH], sfx_cdfm11)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSZOOM], sfx_cdfm01)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSGASP], sfx_cdfm29)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSSKID], sfx_cdfm00)
					MMT_CheckReplaceSfx(player.mo, sfx_drown, sfx_cdfm38)
				elseif player.MMT.MMTchangesfx == "sonic 1"
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSJUMP], sfx_s1a0)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSPUTPUT], sfx_s3kba)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSPUDPUD], sfx_s3kbb)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSSPIN], sfx_s1be)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSZOOM], sfx_s23c)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSGASP], sfx_s1ad)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSSKID], sfx_s1a4)
					MMT_CheckReplaceSfx(player.mo, sfx_drown, sfx_s1b2)
				elseif player.MMT.MMTchangesfx == "sonic 2"
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSJUMP], sfx_s220)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSPUTPUT], sfx_s3kba)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSPUDPUD], sfx_s3kbb)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSSPIN], sfx_s23e)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSSPNDSH], sfx_s260)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSZOOM], sfx_s23c)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSGASP], sfx_s22d)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSSKID], sfx_s224)
					MMT_CheckReplaceSfx(player.mo, sfx_drown, sfx_s232)
				elseif player.MMT.MMTchangesfx == "sonic 3k"
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSJUMP], sfx_s3k62)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSPUTPUT], sfx_s3kba)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSPUDPUD], sfx_s3kbb)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSSPIN], sfx_s3k3c)
					if player.dashspeed -- i'm so sorry for this 
						if player.dashspeed/FRACUNIT <= 23
							if MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSSPNDSH], sfx_s3kab)
							end
						elseif player.dashspeed/FRACUNIT > 23 and player.dashspeed/FRACUNIT < 34
							if MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSSPNDSH], sfx_s3kab1)
								S_StopSoundByID(player.mo, sfx_s3kab)
							end
						elseif player.dashspeed/FRACUNIT > 34 and player.dashspeed/FRACUNIT < 43
							if MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSSPNDSH], sfx_s3kab2)
								S_StopSoundByID(player.mo, sfx_s3kab1)
							end
						elseif player.dashspeed/FRACUNIT > 43 and player.dashspeed/FRACUNIT < 52
							if MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSSPNDSH], sfx_s3kab3)
								S_StopSoundByID(player.mo, sfx_s3kab2)
							end
						elseif player.dashspeed/FRACUNIT > 52 and player.dashspeed/FRACUNIT < 61
							if MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSSPNDSH], sfx_s3kab4)
								S_StopSoundByID(player.mo, sfx_s3kab3)
							end
						elseif player.dashspeed/FRACUNIT > 61 and player.dashspeed/FRACUNIT < 70
							if MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSSPNDSH], sfx_s3kab5)
								S_StopSoundByID(player.mo, sfx_s3kab4)
							end
						elseif player.dashspeed/FRACUNIT >= 70
							if MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSSPNDSH], sfx_s3kab6)
								S_StopSoundByID(player.mo, sfx_s3kab5)
							end
						end
					else
						MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSSPNDSH], sfx_s3kab)
					end
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSZOOM], sfx_s3kb6)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSGASP], sfx_s3k38)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSSKID], sfx_s3k36)
					MMT_CheckReplaceSfx(player.mo, sfx_drown, sfx_s3k3b)
				elseif player.MMT.MMTchangesfx == "knuckles' chaotix"
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSJUMP], sfx_kc5f)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSSPNDSH], sfx_kc6c)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSZOOM], sfx_kc3c)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSSKID], sfx_kc2b)
				elseif player.MMT.MMTchangesfx == "srb2"
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSJUMP], sfx_jump)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSPUTPUT], sfx_putput)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSPUDPUD], sfx_pudpud)				
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSSPIN], sfx_spin)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSSPNDSH], sfx_spndsh)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSZOOM], sfx_zoom)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSGASP], sfx_gasp)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSTHOK], sfx_thok)
					MMT_CheckReplaceSfx(player.mo, skins[player.mo.skin].soundsid[SKSSKID], sfx_skid)
				end
			end
			
			if player.MMT.HitSpring -- overly convoluted workaround, just so i can use postthinkframe; otherwise i'd have to use mobjs.iterate!
				MMT_CheckReplaceSfx(player.MMT.HitSpring, sfx_spring, player.MMT.HitSpring.SFXChanger)
				player.MMT.HitSpring.SFXChanger = nil
				player.MMT.HitSpring = nil
			end
		end
		-- [[ Special Stage Entrance ]] --
		if player.MMT.SSE
			player.mo.flags2 = $ | MF2_DONTDRAW
			if MMTpreserveshield == "off" or player.MMT.ShieldAlreadySaved
				player.powers[pw_shield] = 0
			end
			player.powers[pw_invulnerability] = 0
			player.powers[pw_sneakers] = 0 -- solely because of pointy/fluffy
		end
	end
end)

-- [[ Change Various SFX, for springs ]] --

local function SpringSFXChange(spring, toucher) 
	if netgame return end
	if toucher.type == MT_PLAYER and ((toucher.z + toucher.height > spring.z) and (toucher.z < spring.z + spring.height + 4*FRACUNIT))
		local player = toucher.player
		if player.MMT.MMTchangesfx == "sonic cd"
			spring.SFXChanger = sfx_cdfm08
		elseif player.MMT.MMTchangesfx == "sonic 1"
			spring.SFXChanger = sfx_s1cc
		elseif player.MMT.MMTchangesfx == "sonic 2"
			spring.SFXChanger = sfx_s24c
		elseif player.MMT.MMTchangesfx == "sonic 3k"
			spring.SFXChanger = sfx_s3kb1
		elseif player.MMT.MMTchangesfx == "knuckles' chaotix"
			spring.SFXChanger = sfx_kc2f
		elseif player.MMT.MMTchangesfx == "srb2"
			spring.SFXChanger = sfx_spring
		end
		if player.MMT.MMTchangesfx != "off"
			player.MMT.HitSpring = spring -- overly convoluted workaround, just so i can use postthinkframe; otherwise i'd have to use mobjs.iterate!
		end
	end
end

addHook("MobjCollide", SpringSFXChange, MT_YELLOWSPRING)
addHook("MobjCollide", SpringSFXChange, MT_REDSPRING)
addHook("MobjCollide", SpringSFXChange, MT_BLUESPRING)
addHook("MobjCollide", SpringSFXChange, MT_YELLOWDIAG)
addHook("MobjCollide", SpringSFXChange, MT_REDDIAG)
addHook("MobjCollide", SpringSFXChange, MT_BLUEDIAG)
addHook("MobjCollide", SpringSFXChange, MT_YELLOWHORIZ)
addHook("MobjCollide", SpringSFXChange, MT_REDHORIZ)
addHook("MobjCollide", SpringSFXChange, MT_BLUEHORIZ)
addHook("MobjCollide", SpringSFXChange, MT_SPRINGSHELL)
addHook("MobjCollide", SpringSFXChange, MT_YELLOWSHELL)
addHook("MobjCollide", SpringSFXChange, MT_YELLOWSPRINGBALL)
addHook("MobjCollide", SpringSFXChange, MT_REDSPRINGBALL)


-- [[ Banned Skin HUD Element ]] --
local function skinbannedhud(v, stplyr)
	if stplyr.mo and stplyr.mo.valid and stplyr and stplyr.valid
	and stplyr.MMT.BanSkinNotify
	--v.drawFill(0, 80, 135, 150, V_ORANGEMAP)
	if stplyr.mo.skin == "tailsguy"
		v.fadeScreen(0xFF00, 31)
	else
		v.fadeScreen(0xFF00, 20)
	end
	v.drawString(50, 89, "This skin is currently\x85 banned\x80\.", V_ALLOWLOWERCASE, center)
	v.drawString(60, 104, "\x82\Please choose another skin.", V_ALLOWLOWERCASE, center)
	end
end

hud.add(skinbannedhud)

-- [[ Score-Based Continues / Score Continues ]] --

addHook("IntermissionThinker", function(scorecont) local player = players[0] if player.MMT.ContinueHudDontRender return end
	if not multiplayer and MMTscorecontinues == "on" and player.MMT.scoretocalculate != nil
		if S_IdPlaying(sfx_chchng) or S_IdPlaying(sfx_s3kb0) or S_IdPlaying(sfx_token)
			if not player.MMT.GrantedContinue --and (S_IdPlaying(sfx_chchng) or S_IdPlaying(sfx_s3kb0))
				if player.score - player.MMT.scoretocalculate >= 15000
					mmtdebuggerPrint(player.score - player.MMT.scoretocalculate)
					mmtdebuggerPrint("seY")
					player.continues = $ + 1
					player.MMT.GotContinue = true
					--S_StartSound(nil, sfx_s3kac, player)
				else
					mmtdebuggerPrint(player.score - player.MMT.scoretocalculate)
					player.MMT.GotContinue = false
				end
			end
			player.MMT.GrantedContinue = true
		end
	end
end)

local function ContinueHud(v, player) if multiplayer return end 
	local player = player
	if not player
		player = displayplayer -- this also doesn't allow players for some reason >:( plz stjr
	end
	if player.MMT.ContinueHudDontRender return end
	
	if not player.MMT.GotContinue
		v.timer = nil -- this will force reset everything
	else
		if v.timer == nil 
			v.timer = 0
			v.GravityY = -40
			v.bounces = 0
			v.momy = 20
		end
		v.timer = $ + 1
		local contsprite = v.getSprite2Patch(player.MMT.skincont, SPR2_XTRA, false, C)
		local COLOR = v.getColormap("sonic", player.MMT.colorcont)
		
		if v.bounces < 3 -- bounce code heavily inspired from SRB2's y_inter.c 's emeraldbounce code
			v.GravityY = $ + v.momy
			v.momy = $ + 1
			if v.GravityY > 161
				S_StartSound(nil, sfx_tink, player)
				v.bounces = $ + 1
				v.momy = -($/2)
				v.GravityY = 161
			end
		end
		
		v.YOffset = (5*(player.continues-1))
		if player.continues > 5
			v.YOffset = (5*(5-1))
		end
		
		
		mmtdebuggerPrint(v.GravityY .. " " .. v.bounces .. " " .. v.momy, true)
		v.draw(267, v.GravityY - v.YOffset, contsprite, nil, COLOR) -- do not snap to any direction, intermission hud doesn't do that. also, the Y to aim for is 161
	end
	
	-- [ render continues ] --
	local continuesrender = player.continues
	local extras
	if continuesrender > 4
		continuesrender = 4
		extras = player.continues - 4
	end
	if player.MMT.GotContinue
		continuesrender = $ - 1
		if extras
			continuesrender = $ + 1
			extras = player.continues - 5
		end
	end
	
	for i=continuesrender,1,-1 do 
		local contsprite = v.getSprite2Patch(player.MMT.skincont, SPR2_XTRA, false, C)
		local COLOR = v.getColormap("sonic", player.MMT.colorcont)
		local PlacementY = 161 + 5
		
		v.draw(267, PlacementY - (5*i), contsprite, nil, COLOR)
		if extras v.drawString(267+14, 156, "+ " +extras) end
	end
end

hud.add(ContinueHud, "intermission")

-- [[ Emerald Token to Rings ]] --

addHook("TouchSpecial", function(emeraldtoken, player)
	if MMTclassicspecialstage != "50 rings, no tokens" or MMTclassicspecialstage != "50 rings, keep tokens"
		if not emeraldtoken.CollectedTime and MMTtokentorings == "on" and All7Emeralds(emeralds)
			mmtdebuggerPrint("Emerald Token collected, granted 50 rings (hopefully!)")
			player.player.MMT.TokenRingCollected = true
			emeraldtoken.CollectedTime = true
			P_KillMobj(emeraldtoken)
			return true
		end
	else
		mmtdebuggerPrint("Emerald Token is normal!")
		return false 
	end
end, MT_TOKEN)

--[[addHook("TouchSpecial", function(emeraldtoken, toucher) -- replacement token code
	if MMTtokenbehavior == "on"
		if toucher.player.bot return end
		MMTtokens = $ + 1
		MMTtokenlist = $ + emeraldtoken.health
		P_AddPlayerScore(toucher.player, 1000)
		S_StartSound(toucher, sfx_token)
		P_KillMobj(emeraldtoken)
		return true
	else
		mmtdebuggerPrint("Emerald Token is normal!")
		return false 
	end

end, MT_TOKEN)]]

--addHook("MobjThinker", function(token) -- didn't work with mobjspawn, because it happened before the health was overwriten with the correct values. thanks again stjr 
--	if (MMTtokenlist & token.health)
--		mmtdebuggerPrint("i existed once, but never again, dying now")
--		P_KillMobj(token)
--	end
--end, MT_TOKEN)

-- [[ Classic Special Stage ]] --

addHook("MobjThinker", function(cssemeraldtoken) if multiplayer return end	
	if MMTclassicspecialstage != "50 rings, no tokens" 
		cssemeraldtoken.flags = $1 | MF_SPECIAL
		cssemeraldtoken.flags2 = $ & ~MF2_DONTDRAW 
		return 
	end
	if G_EnoughPlayersFinished() or All7Emeralds(emeralds)
		cssemeraldtoken.flags = $1 | MF_SPECIAL
		cssemeraldtoken.flags2 = $ & ~MF2_DONTDRAW
		--mmtdebuggerPrint("all is good now :\)")
	else
		cssemeraldtoken.flags2 = $1 | MF2_DONTDRAW
		cssemeraldtoken.flags = $ & ~MF_SPECIAL
		--mmtdebuggerPrint("help")
	end
end, MT_TOKEN)

-- [[ Disable Super Hover ]] --
addHook("JumpSpinSpecial", function(player)
	if player.powers[pw_super] and player.charability == CA_THOK and (player.MMT.MMTsuperhover == "off")
		return true
	end
end)


addHook("PlayerSpawn", function(player) -- MMTtokenlist = 0
	if not player.MMT player.MMT = {} end
	if not player.MMTHud player.MMTHud = {} end
	player.MMT.FakeShield = nil
	
	if player.MMT.SSE
		player.mo.flags2 = $ & ~MF2_DONTDRAW
		player.MMT.SSE = false
	end
	
	player.MMT.BanSkinNotify = false
	
	player.MMT.EnteredCS = false
	player.MMT.ContinueHudDontRender = nil
	
	player.MMT.loadtimer = 0
	
	
	
	player.MMT.GrantedContinue = false
	player.MMT.GotContinue = false
	
	player.MMT.MenuActive = false
	
	player.MMT.waittogetdown = false
	player.MMT.istailsball = false
	player.MMT.initialfall = false
	player.MMT.beginthebopping = false
	player.MMT.allowreflight = false
	player.MMT.usedreflight = false
	
	player.MMT.TokenRingCollected = false
	
	player.MMT.beginfade = false
	
	player.MMT.MusicVolume = 100
	
	-- [[ Preserve Shield ]] --
	if player.MMT.ShieldAlreadySaved and MMTpreserveshield == "on" and not G_IsSpecialStage(gamemap)
		if (player.MMT.ShieldLevelName == mapheaderinfo[gamemap].lvlttl) or player.MMT.MMTpreserveshieldzones == "on"
			player.powers[pw_shield] = player.MMT.ShieldAfterLevel
			mmtdebuggerPrint("Shield spawned is: " +player.MMT.ShieldAfterLevel)
			P_SpawnShieldOrb(player)
		end
		player.MMT.ShieldAlreadySaved = false
	end
end)

-- [[ Total Rings / Count Rings ]] -- 

local stagerings = 0
local stageringscollected = 0
local currentstage 

addHook("MapChange", do
	stagerings = 0
	stageringscollected = 0
end)


addHook("MobjSpawn", function(mobj)
	stagerings = $ + 1
end, MT_RING)

addHook("MobjSpawn", function(mobj)
	stagerings = $ + 1
end, MT_COIN)

addHook("MobjSpawn", function(mobj)
	stagerings = $ + 10
end, MT_RING_BOX)


addHook("MobjDeath", function(mobj)
	stageringscollected = $ + 1
end, MT_RING)

addHook("MobjDeath", function(mobj)
	stageringscollected = $ + 1
end, MT_COIN)

addHook("MobjDeath", function(mobj)
	stageringscollected = $ + 10
end, MT_RING_BOX)

local function countrings(v, player)
	local vlayer = player
	if not vlayer
		vlayer = displayplayer -- just because the scores function doesn't allow for player for some reason >:(
	end
	local yringoffset = 0
	local xringoffset = 0
	local xringoffset2 = 0
	if vlayer.mo and vlayer.mo.valid and vlayer and vlayer.valid 
	and (vlayer.MMT.MMTtotalrings == "on") and G_PlatformGametype() and not (mapheaderinfo[gamemap].typeoflevel & TOL_NIGHTS) and not G_IsSpecialStage(gamemap)
	and stagerings > 0
		if modifiedgame
		yringoffset = $ - 32
		end
		
		if vlayer.rings < 10
		elseif vlayer.rings >= 10 and vlayer.rings < 100
			xringoffset = $ + 8
		elseif vlayer.rings >= 100 and vlayer.rings < 1000
			xringoffset = $ + 16
		elseif vlayer.rings >= 1000
			xringoffset = $ + 24
		end
		
		if stagerings < 10
		elseif stagerings >= 10 and stagerings < 100
			xringoffset2 = $ + 8
		elseif stagerings >= 100 and stagerings < 1000
			xringoffset2 = $ + 16
		elseif stagerings >= 1000
			xringoffset2 = $ + 24
		end
		
		if not v.Timer
			v.Timer = 0
		end
		v.Timer = $ + 1
		
		if v.Timer > 23
			v.Timer = 0
		end

		local RingSprite, Flipper = v.getSpritePatch("RING", v.Timer)
		if Flipper == true
			v.FLIPFLAG = V_FLIP
		else
			v.FLIPFLAG = nil
		end
		
		local ColorMap
		
		if not multiplayer and vlayer.rings >= stagerings
			ColorMap = V_GREENMAP|V_PERPLAYER
		elseif not multiplayer and stageringscollected > vlayer.rings 
			ColorMap = V_REDMAP|V_PERPLAYER
		else
			ColorMap = V_PERPLAYER
		end

		v.drawScaled(147*FRACUNIT, (171+yringoffset)*FRACUNIT, FRACUNIT/2, RingSprite, v.FLIPFLAG) -- X value used to be 151
		v.drawString(161, 160+yringoffset, "-", V_ALLOWLOWERCASE, left)
		
		v.drawString(172, 160+yringoffset, vlayer.rings, ColorMap)
		v.drawString(180+xringoffset, 160+yringoffset, "/", ColorMap)
		v.drawString(188+xringoffset, 160+yringoffset, stagerings, ColorMap)
		if not multiplayer and vlayer.rings!= stageringscollected
			if ColorMap == V_REDMAP|V_PERPLAYER
				v.AddOrMinus = ""
			else
				ColorMap = V_GREENMAP|V_PERPLAYER
				v.AddOrMinus = "+"
			end
			v.drawString(198+(xringoffset+xringoffset2), 160+yringoffset, "(" .. v.AddOrMinus .. vlayer.rings - stageringscollected .. ")", ColorMap) 
		end
	end
end

hud.add(countrings, "scores")

-- [[ Super Shield Mobjs ]] --

addHook("MobjThinker", function(shield) -- there are so many .valid checks in here, when one goes down i need to patch another >:( don't whine at me for being lazy, blame the game for whining at me
	if not shield.valid return end
	local player
	if shield and shield.valid
		if not shield.target or not shield.target.valid
			if shield.friendly and shield.friendly.valid 
				shield.friendly.TerminateSelf = true 
			end
			P_RemoveMobj(shield)
			mmtdebuggerPrint("target not found, delet this")
		end

		if shield.valid and shield.TerminateSelf 
		--or (shield.target.valid and shield.target.player.playerstate == PST_DEAD)
			if shield.friendly and shield.friendly.valid
				shield.friendly.TerminateSelf = true
			end
			P_RemoveMobj(shield)
			mmtdebuggerPrint("delet this")
		end
		
		if shield.valid and not shield.TerminateSelf	
			local player = shield.target.player
			if not shield.printed
				mmtdebuggerPrint("shield has been summoned")
				shield.printed = true
			end
		
			-- [ flip ] -- 
			if (player.mo.eflags & MFE_VERTICALFLIP)
				shield.flags2 = $1 | MF2_OBJECTFLIP
				shield.eflags = $1 | MFE_VERTICALFLIP
			else
				shield.flags2 = $ & ~MF2_OBJECTFLIP
				shield.eflags = $ & ~MFE_VERTICALFLIP
			end
			
			P_SetScale(shield, FixedMul(player.mo.scale, player.shieldscale))
			
			-- [ flop ] --
			if (shield.eflags & MFE_VERTICALFLIP)
				shield.destinationz = player.mo.z + (player.mo.height - shield.height + FixedDiv(P_GetPlayerHeight(player) - player.mo.height, 3*FRACUNIT)) - FixedMul(2*FRACUNIT, player.mo.scale)
			else
				shield.destinationz = player.mo.z - (FixedDiv(P_GetPlayerHeight(player) - player.mo.height, 3*FRACUNIT)) + FixedMul(2*FRACUNIT, player.mo.scale)
			end
			
			P_TeleportMove(shield, player.mo.x, player.mo.y, shield.destinationz)
			
			
			if not shield.summonedfriend
				shield.friendly = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_FAKE_SHIELD2)
				shield.friendly.them = player.mo
				shield.friendly.target = shield
				shield.summonedfriend = true
			end
			
			if shield.PlayerShield != player.powers[pw_shield] and shield.SetState
				shield.SetState = false
				mmtdebuggerPrint("unset false shield, prepping for new one")
			end
			
			--[[if (player.pflags & PF_SHIELDABILITY) and not shield.SetShieldAbility
				shield.SetShieldAbility = true
				shield.changesecondstate = true
				mmtdebuggerPrint("player used shield ability")
			end
			
			if (player.pflags & ~PF_SHIELDABILITY) and shield.SetShieldAbility
				shield.SetShieldAbility = false
				mmtdebuggerPrint("shield ability ended, unsetting shield.SetShieldAbility")
			end]]
			
			if not shield.DontRender
				if not shield.SetState
					if player.powers[pw_shield] == SH_WHIRLWIND
						shield.state = S_WIND1
					elseif player.powers[pw_shield] == SH_PITY
						shield.state = S_PITY1
					elseif player.powers[pw_shield] == SH_ARMAGEDDON -- create seperate object for S_ARMF1
						shield.state = S_ARMA1
					elseif player.powers[pw_shield] == SH_PINK -- change skincolor to pink in seperate function
						shield.state = S_PITY1
					elseif player.powers[pw_shield] == SH_ELEMENTAL -- create seperate object for S_ELEMF1
						shield.state = S_ELEM1
					elseif player.powers[pw_shield] == SH_ATTRACT
						shield.state = S_MAGN1
					elseif player.powers[pw_shield] == SH_FLAMEAURA
						shield.state = S_FIRS1
					elseif player.powers[pw_shield] == SH_BUBBLEWRAP
						shield.state = S_BUBS1
					elseif player.powers[pw_shield] == SH_THUNDERCOIN
						shield.state = S_ZAPS1
					elseif player.powers[pw_shield] == SH_FORCE
						shield.state = S_FORC11
					elseif player.powers[pw_shield] == SH_FORCE|1
						shield.state = S_FORC1
					elseif player.powers[pw_shield] == SH_NONE
						shield.state = S_INVISIBLE
					end
					
					mmtdebuggerPrint("set new state")
					if player.powers[pw_shield] == SH_PINK
						shield.color = SKINCOLOR_PINK
						shield.colorized = true
					else
						shield.color = SKINCOLOR_NONE
						shield.colorized = false
					end
					shield.changesecondstate = true
					shield.SetState = true
					shield.PlayerShield = player.powers[pw_shield]
				end
			else
				shield.state = S_INVISIBLE
			end
			
			if not (player.powers[pw_super] or player.powers[pw_invulnerability]) and not shield.DontRender
				mmtdebuggerPrint("player no longer super, disabling render")
				shield.DontRender = true
			end
		end
	end
end, MT_FAKE_SHIELD)

addHook("MobjThinker", function(shield) 
	if not shield.valid return end

	local friend = shield.target 

	if shield.TerminateSelf or not shield.them.valid or shield.them.player.playerstate == PST_DEAD
		P_RemoveMobj(shield)
		mmtdebuggerPrint("delet this 2")
	end

	local player
	if shield.valid and shield.them and shield.them.valid
	player = shield.them.player
	end


	if shield.valid and friend and not shield.TerminateSelf
		
		if (friend.eflags & MFE_VERTICALFLIP)
			shield.flags2 = $1 | MF2_OBJECTFLIP
			shield.eflags = $1 | MFE_VERTICALFLIP
		else
			shield.flags2 = $ & ~MF2_OBJECTFLIP
			shield.eflags = $ & ~MFE_VERTICALFLIP
		end
	
	
		P_TeleportMove(shield, friend.x, friend.y, friend.z)
		
		if not shield.printed
			mmtdebuggerPrint("shield2 has been summoned")
			shield.printed = true
		end
		
		if not friend.DontRender
			if friend.changesecondstate
				if player.powers[pw_shield] == SH_ARMAGEDDON -- create seperate object for S_ARMF1
					shield.state = S_ARMF1
				elseif player.powers[pw_shield] == SH_ELEMENTAL -- create seperate object for S_ELEMF1
					shield.state = S_ELEMF1
				elseif player.powers[pw_shield] == SH_FLAMEAURA
					shield.state = S_FIRSB1
				elseif player.powers[pw_shield] == SH_BUBBLEWRAP
					shield.state = S_BUBSB1
				elseif player.powers[pw_shield] == SH_THUNDERCOIN
					shield.state = S_ZAPSB1
				else
					shield.state = S_INVISIBLE
				end
				if friend.SetShieldAbility
					if player.powers[pw_shield] == SH_FLAMEAURA
						shield.state = mobjinfo[MT_FLAMEAURA_ORB].painstate -- idk how any of this shit works, i can't even find the state used for the shields, stjr plz 
					--elseif
					
					end
				end
				
				
				mmtdebuggerPrint("change second state")
				friend.changesecondstate = false
			end
		else
			shield.state = S_INVISIBLE
		end
	end
end, MT_FAKE_SHIELD2)

-- [[ aggressive pizza ]] --

addHook("MobjThinker", function(aggressive)
	if aggressive.valid and aggressive.target and aggressive.target.valid
		if not aggressive.initializedtime
			aggressive.shadowscale = 3*FRACUNIT
			if aggressive.sizeoption == "large"
				aggressive.speed = 44*FRACUNIT
				aggressive.scale = FRACUNIT + FRACUNIT*(2/4)
			elseif aggressive.sizeoption == "medium"
				aggressive.speed = 39*FRACUNIT
				aggressive.scale = FRACUNIT
			elseif (aggressive.sizeoption == "small" or aggressive.sizeoption == "personal")
				aggressive.speed = (28*3)*FRACUNIT
				aggressive.scale = FRACUNIT/4
			elseif (aggressive.sizeoption == "xl" or aggressive.sizeoption == "extra large" or aggressive.sizeoption == "extralarge")
				aggressive.speed = 49*FRACUNIT
				aggressive.scale = FRACUNIT*2
			else
				aggressive.speed = (36*2)*FRACUNIT
				aggressive.scale = FRACUNIT/2
			end
			aggressive.initializedtime = true
		end
		if not aggressive.rollanglesprite
		aggressive.rollanglesprite = 0
		end
		aggressive.rollanglesprite = $ + ANG15
		aggressive.rollangle = aggressive.rollanglesprite

		if (R_PointToDist2(aggressive.target.x, aggressive.target.y, aggressive.x, aggressive.y))/FRACUNIT < 35 and (abs((aggressive.target.z - aggressive.z)/FRACUNIT) < 25) and not (aggressive.target.player and aggressive.target.player.playerstate == PST_DEAD)
			if aggressive.target.player.rings > 0
				P_PlayerRingBurst(aggressive.target.player)
				aggressive.target.player.rings = 0
				S_StartSound(aggressive.target, sfx_s3kb9)
			end
			--aggressive.target.health = 0
			P_DamageMobj(aggressive.target, aggressive, aggressive, 666, DMG_INSTAKILL)
			S_StartSound(aggressive.target, sfx_pizkil)
			if aggressive.state == S_THEPIZZAISEXPLOSIVE
				local Explode = P_SpawnMobj(aggressive.target.x, aggressive.target.y, aggressive.target.z, MT_BOSSEXPLODE)
				S_StartSound(explode, sfx_bkpoof)
				Explode.destscale = 3*FRACUNIT
				Explode.scalespeed = FRACUNIT/7
			end
			aggressive.target = nil
		end

		if aggressive.target
			A_HomingChase(aggressive, aggressive.speed, 0)
		else
			P_RemoveMobj(aggressive)
		end
	else
		P_RemoveMobj(aggressive)
	end
end, MT_AGGRESSIVE)

-- [[ Classic Shields ]] --

local function ChangeShield(classicshield)
	if MMTclassicshields == "on"
		classicshield.prevname = classicshield.type
		if classicshield.type == MT_ATTRACT_BOX
			classicshield.type = MT_THUNDERCOIN_BOX
		elseif classicshield.type == MT_ATTRACT_GOLDBOX
			classicshield.type = MT_THUNDERCOIN_GOLDBOX

		elseif classicshield.type == MT_FORCE_BOX
			classicshield.type = MT_BUBBLEWRAP_BOX
		elseif classicshield.type == MT_FORCE_GOLDBOX
			classicshield.type = MT_BUBBLEWRAP_GOLDBOX	

		elseif classicshield.type == MT_WHIRLWIND_BOX
			classicshield.type = MT_THUNDERCOIN_BOX
		elseif classicshield.type == MT_WHIRLWIND_GOLDBOX
			classicshield.type = MT_THUNDERCOIN_GOLDBOX

		elseif classicshield.type == MT_ARMAGEDDON_BOX
			classicshield.type = MT_FLAMEAURA_BOX
		elseif classicshield.type == MT_ARMAGEDDON_GOLDBOX
			classicshield.type = MT_FLAMEAURA_GOLDBOX	

		elseif (classicshield.type == MT_ELEMENTAL_BOX or classicshield.type == MT_ELEMENTAL_GOLDBOX)
			if mapheaderinfo[gamemap].lvlttl == "Deep Sea" or mapheaderinfo[gamemap].lvlttl == "Greenflower" or mapheaderinfo[gamemap].lvlttl == "Egg Rock"
				mmtdebuggerPrint("Map is water stage, and elemental shield detected, it's now bubble shield.")
				classicshield.ChangeType = MT_BUBBLEWRAP_BOX
			elseif mapheaderinfo[gamemap].lvlttl == "Red Volcano"
				classicshield.ChangeType = MT_FLAMEAURA_BOX
				mmtdebuggerPrint("Map is lava stage, and elemental shield detected, it's now fire shield.")
			else
				mmtdebuggerPrint("Map name not detected, choosing random shield for elemental shield.")
				if P_RandomByte() < 127
					classicshield.ChangeType = MT_BUBBLEWRAP_BOX
				else
					classicshield.ChangeType = MT_FLAMEAURA_BOX
				end
			end

			if classicshield.type == MT_ELEMENTAL_GOLDBOX
				if classicshield.ChangeType == MT_BUBBLEWRAP_BOX
					classicshield.ChangeType = MT_BUBBLEWRAP_GOLDBOX
				else
					classicshield.ChangeType = MT_FLAMEAURA_GOLDBOX
				end
			end
			classicshield.type = classicshield.ChangeType
		end
	end
end

local function OnMonitorHit(classicshield, inflictor, source)
	mmtdebuggerPrint("i'm hit")
	if classicshield.prevname == nil
		mmtdebuggerPrint("not a classicshield monitor")
	else
		mmtdebuggerPrint("yowch, my previous name was:" .. classicshield.prevname)
		if source
		mmtdebuggerPrint("source")
		source.player.ShieldPrevname = classicshield.prevname
		--[[else
		mmtdebuggerPrint("inflictor")
		inflictor.player.ShieldPrevname = classicshield.prevname]]
		end
	end
end

local function RevertClassicShieldFunc(classicshield)
	if MMTclassicshields == "off"
		if classicshield.prevname != nil
			classicshield.type = classicshield.prevname
		end
	end
end

addHook("MobjThinker", ChangeShield, MT_ATTRACT_BOX)
addHook("MobjThinker", ChangeShield, MT_ATTRACT_GOLDBOX)
addHook("MobjThinker", ChangeShield, MT_FORCE_BOX)
addHook("MobjThinker", ChangeShield, MT_FORCE_GOLDBOX)
addHook("MobjThinker", ChangeShield, MT_WHIRLWIND_BOX)
addHook("MobjThinker", ChangeShield, MT_WHIRLWIND_GOLDBOX)
addHook("MobjThinker", ChangeShield, MT_ARMAGEDDON_BOX)
addHook("MobjThinker", ChangeShield, MT_ARMAGEDDON_GOLDBOX)
addHook("MobjThinker", ChangeShield, MT_ELEMENTAL_BOX)
addHook("MobjThinker", ChangeShield, MT_ELEMENTAL_GOLDBOX)

addHook("MobjThinker", RevertClassicShieldFunc, MT_FLAMEAURA_BOX)
addHook("MobjThinker", RevertClassicShieldFunc, MT_BUBBLEWRAP_BOX)
addHook("MobjThinker", RevertClassicShieldFunc, MT_THUNDERCOIN_BOX)
addHook("MobjThinker", RevertClassicShieldFunc, MT_FLAMEAURA_GOLDBOX)
addHook("MobjThinker", RevertClassicShieldFunc, MT_BUBBLEWRAP_GOLDBOX)
addHook("MobjThinker", RevertClassicShieldFunc, MT_THUNDERCOIN_GOLDBOX)

addHook("MobjDeath", OnMonitorHit, MT_THUNDERCOIN_BOX)
addHook("MobjDeath", OnMonitorHit, MT_BUBBLEWRAP_BOX)
addHook("MobjDeath", OnMonitorHit, MT_FLAMEAURA_BOX)
addHook("MobjDeath", OnMonitorHit, MT_THUNDERCOIN_GOLDBOX)
addHook("MobjDeath", OnMonitorHit, MT_BUBBLEWRAP_GOLDBOX)
addHook("MobjDeath", OnMonitorHit, MT_FLAMEAURA_GOLDBOX)