-- &
local thosbutons = {
	"up","down","left","right",
	"spin","jump",
	"attack","nextwep","prevwep",
	"custom1","custom2","custom3",
	"moving","hmoving","vmoving",
	"attacknormal", "camleft", "camright", "tossflag" -- birbhorse additions
}
for k,v in ipairs(thosbutons) do
	thosbutons[v] = true
end

--NOTE: this has been modified specifically for MMT, all "p.bt..." variables have been converted to "p.MMTbt..." so it doesn't conflict with any other mod that uses this system

rawset(_G, "pbtDaemonMMT", function(p)
	--if deepcopy == nil then error("deepcopy missing", 2) end
	// Does a cool thing
	local cmdbt = p.cmd.buttons
	local cmdfm = p.cmd.forwardmove
	local cmdsm = p.cmd.sidemove
	
	if not p.MMTbttimestamp then p.MMTbttimestamp = -1 end
	
	if p.MMTbttimestamp ~= leveltime then
		// Prevent this from running more than once if multiple daemons are in place (why would there be more than one)
		
		if not p.MMTbt then
			p.MMTbt = {}
			p.MMTbtx = {}
			p.MMTbtprev = {}
			for _,v in ipairs(thosbutons) do
				p.MMTbt[v] = 0
				p.MMTbtprev[v] = 0
				p.MMTbtx[v] = 0
			end
		end
		local sens = 20
		
		for i = 1,#thosbutons do
			local ix = thosbutons[i]
			p.MMTbtprev[ix] = p.MMTbt[ix]
		end // Old values
		
		p.MMTbt.up = cmdfm > sens and $+1 or 0
		p.MMTbt.down = cmdfm < -sens and $+1 or 0
		p.MMTbt.right = cmdsm > sens and $+1 or 0
		p.MMTbt.left = cmdsm < -sens and $+1 or 0
		
		if p.MMTbt.up > 0 and p.MMTbt.down > 0 then p.MMTbt.up = 0; p.MMTbt.down = 0 end
		if p.MMTbt.right > 0 and p.MMTbt.left > 0 then p.MMTbt.right = 0; p.MMTbt.left = 0 end
		
		p.MMTbt.spin = (cmdbt & BT_USE) > 0 and $+1 or 0
		p.MMTbt.jump = (cmdbt & BT_JUMP) > 0 and $+1 or 0
		p.MMTbt.attack = (cmdbt & BT_ATTACK) > 0 and $+1 or 0
		p.MMTbt.attacknormal = (cmdbt & BT_FIRENORMAL) > 0 and $+1 or 0
		
		p.MMTbt.tossflag = (cmdbt & BT_TOSSFLAG) > 0 and $+1 or 0
		
		p.MMTbt.camleft = (cmdbt & BT_CAMLEFT) > 0 and $+1 or 0
		p.MMTbt.camright = (cmdbt & BT_CAMRIGHT) > 0 and $+1 or 0
		
		p.MMTbt.nextwep = (cmdbt & BT_WEAPONNEXT) > 0 and $+1 or 0
		p.MMTbt.prevwep = (cmdbt & BT_WEAPONPREV) > 0 and $+1 or 0
		
		p.MMTbt.custom1 = (cmdbt & BT_CUSTOM1) > 0 and $+1 or 0
		p.MMTbt.custom2 = (cmdbt & BT_CUSTOM2) > 0 and $+1 or 0
		p.MMTbt.custom3 = (cmdbt & BT_CUSTOM3) > 0 and $+1 or 0
		
		p.MMTbt.hmoving = (p.MMTbt.left or p.MMTbt.right) and $~=nil and $+1 or 0
		p.MMTbt.vmoving = (p.MMTbt.up or p.MMTbt.down) and $~=nil and $+1 or 0
		p.MMTbt.moving = (p.MMTbt.up or p.MMTbt.down or p.MMTbt.left or p.MMTbt.right) and $~=nil and $+1 or 0
		
		local sum = 0
		local sidesum = 0
		
		if cmdsm < 0 then sidesum = 45 end
		if cmdsm > 0 then sidesum = -45 end
		
		if cmdfm == 0 then sidesum = $*2 end
		if cmdfm < 0 then sidesum = -sidesum;sum = 180 end
		
		local res = sum + sidesum
		--return FixedAngle( res*FRACUNIT )
		p.MMTbt.dirAsAngle = FixedAngle( res*FRACUNIT )
		
		p.MMTbttimestamp = leveltime
	end
end)

/*
&pbt usage:

p.MMTbt holds button held time
p.MMTbtprev holds the value in the previous frame for that respective button

Want to check if a player is holding jump?
p.MMTbt.jump > 0

Want to check for a tap on the jump button (not held, but the instant it is pressed)?
p.MMTbt.jump == 1

Want to check when jump was just released?
p.MMTbt.jump < p.MMTbtprev.jump
normally btprev is smaller than bt when the button is held, so if it is larger, it must have been released


or use one of the shortcut functions like a baby
*/

local function checks(p, key)
	if not p then error("Bad argument #1 (expected player)", 3) end
	if not p.MMTbt then
		error("Are you running pbtDaemonMMT()?", 3)
	end
	if key == nil then error("Key argument is nil", 3) end
	if not thosbutons[key] then
		error("Button "..key.." not in the list", 3)
	end
end

rawset(_G, "PBTMMT_Up", function(p, key) checks(p,key)
	return not p.MMTbt[key] // If 0, true
end)
rawset(_G, "PBTMMT_Down", function(p, key) checks(p,key)
	return not not p.MMTbt[key] // If >0, true
end)

rawset(_G, "PBTMMT_JustPressed", function(p, key) checks(p,key)
	// Button has just been pressed after not being pressed
	return p.MMTbt[key] and p.MMTbtprev[key] == 0
end)
rawset(_G, "PBTMMT_JustReleased", function(p, key) checks(p,key)
	// Button is no longer being pressed after being pressed press press press press press press press press press press 
	return p.MMTbt[key] < p.MMTbtprev[key]
end)